/* --------------------------------------------------------------- */
/* Licensed Materials - Property of IBM                            */
/* 5724-S84                                                        */
/* (C) Copyright IBM Corp. 2008       All Rights Reserved          */
/* US Government Users Restricted Rights - Use, duplication or     */
/* disclosure restricted by GSA ADP Schedule Contract with         */
/* IBM Corp.                                                       */
/* --------------------------------------------------------------- */
/* PROLOG END TAG zYx                                              */
#include <stdio.h>
#include <stdlib.h>
#include <spu_mfcio.h>
#include "triple_buffer.h"

//#define USE_TIMER       1

#ifdef USE_TIMER
#include <spu_timer.h>
#endif /* USE_TIMER */


#ifdef COMBINED_CASE
#define INVARIANT(X)
#define AUXILIARY(X)
#else
#define INVARIANT(X) X
#define AUXILIARY(X) X
#endif
#define REQUIRED_INVARIANT(X) X


AUXILIARY(int the_process_buf);
AUXILIARY(unsigned int tag1);
AUXILIARY(unsigned int tag2);
AUXILIARY(unsigned int tag3);
AUXILIARY(unsigned int the_get_tag);
AUXILIARY(unsigned int the_put_tag);
AUXILIARY(int the_get_buf);
AUXILIARY(int the_put_buf);


/* here we define the data buffer in local memory to hold the data we
 * DMA in.  We will DMA CHUNK_SIZE elements at a time.  */
float local_buffers[3][CHUNK_SIZE] __attribute__ ((aligned (128)));

/* control block structure */
control_block_t control_block __attribute__ ((aligned (128)));

#ifdef NO_SLICE 

/*
 * assume that buf is properly aligned, size is multiple of 16
 */ 
void process_data_simd (float* buf, unsigned int size)
{
	unsigned int i;
	vec_float4 *vbuf;
	vec_float4 v1 = (vec_float4) {1.0f, 1.0f, 1.0f, 1.0f};
	vbuf = (vec_float4*) buf;
	
	for (i = 0; i < (size / 4); i++)
	{
#ifdef COMBINED_CASE
		__CPROVER_k_induction_hint(0,0,1,0);
#endif
#ifdef SPLIT_CASE
		__CPROVER_k_induction_hint(1,1,1,0);
#endif
		
		INVARIANT(assert(buf == local_buffers[0]));
		INVARIANT(assert(buf == vbuf));
		INVARIANT(assert(size == CHUNK_SIZE));
		INVARIANT(assert(tag1 == 0));
		INVARIANT(assert(tag2 == 1));
		INVARIANT(assert(tag3 == 2));
		INVARIANT(assert_mfc_getb (local_buffers[1], 0, CHUNK_SIZE * sizeof(float), 1, 0, 0));
		INVARIANT(assert_mfc_getb (local_buffers[2], 0, CHUNK_SIZE * sizeof(float), 2, 0, 0));
		INVARIANT(dma_free_region(vbuf, CHUNK_SIZE*sizeof(float)));

#ifdef USE_SIMD
		vbuf[i] = spu_add (vbuf[i], v1); 
#endif
	}
}


/*
 * assume that buf is properly aligned, size is multiple of 16
 */ 
void process_data_simd2 (float* buf, unsigned int size)
{
	unsigned int i;
	vec_float4 *vbuf;
	vec_float4 v1 = (vec_float4) {1.0f, 1.0f, 1.0f, 1.0f};
	vbuf = (vec_float4*) buf;
	
	for (i = 0; i < (size / 4); i++)
	{
#ifdef COMBINED_CASE
		assert(0);
#endif
#ifdef SPLIT_CASE
		__CPROVER_k_induction_hint(1,1,1,0);
#endif
				
		INVARIANT(assert(buf == local_buffers[the_process_buf]));
		INVARIANT(assert(the_get_buf != the_put_buf));
		INVARIANT(assert(the_get_buf != the_process_buf));
		INVARIANT(assert(the_put_buf != the_process_buf));
		INVARIANT(assert(vbuf == buf));
		INVARIANT(assert(tag1 == 0));
		INVARIANT(assert(tag2 == 1));
		INVARIANT(assert(tag3 == 2));
		INVARIANT(assert(size == CHUNK_SIZE));	  
		INVARIANT(assert(0 <= the_get_buf && the_get_buf <= 2));
		INVARIANT(assert(0 <= the_put_buf && the_put_buf <= 2));
		INVARIANT(assert(0 <= the_process_buf && the_process_buf <= 2));
		INVARIANT(assert_mfc_getb (local_buffers[the_get_buf], 0, CHUNK_SIZE * sizeof(float),
						 the_get_tag, 0, 0));
		INVARIANT(assert_mfc_getb (local_buffers[the_put_buf], 0, CHUNK_SIZE * sizeof(float),
						 the_put_tag, 0, 0));
		INVARIANT(dma_free_region(vbuf, CHUNK_SIZE*sizeof(float)));

#ifdef USE_SIMD
		vbuf[i] = spu_add (vbuf[i], v1); 
#endif
	}
}
#endif


int spu_main(unsigned long long speid __attribute__ ((unused)),
         unsigned long long argp, 
         unsigned long long envp __attribute__ ((unused)))
{
  unsigned int tags[3];
  unsigned long long in_addr, out_addr;
  unsigned int i, num_chunks;
  int put_buf, get_buf, process_buf;

#ifdef USE_TIMER
  uint64_t start, time_working;
  spu_slih_register (MFC_DECREMENTER_EVENT, spu_clock_slih);
  spu_clock_start();
  start = spu_clock_read();
#endif /* USE_TIMER */


  /* First, we reserve two MFC tags for use with double buffering */
  tags[0] = TAG_RESERVE(0);
  tags[1] = TAG_RESERVE(1);
  tags[2] = TAG_RESERVE(2);

  if ((tags[0] == MFC_TAG_INVALID) || (tags[1] == MFC_TAG_INVALID) || (tags[2] == MFC_TAG_INVALID))
  {
    printf ("SPU ERROR, unable to reserve tag\n");
    return 1;
  }
  /* DMA the control block information from system memory */
  mfc_get (&control_block, argp, sizeof (control_block_t), tags[0], 0, 0);
  HAVOC_CONTROL_BLOCK(control_block);
  mfc_write_tag_mask (1 << tags[0]);
  mfc_read_tag_status_all ();

  num_chunks = control_block.num_elements_per_spe/CHUNK_SIZE;

  /* initialize addr, out_addr */
  in_addr = control_block.in_addr;
  out_addr = control_block.out_addr;

  /* Issue the first DMA */
  mfc_get (local_buffers[0], in_addr, CHUNK_SIZE * sizeof(float),
      tags[0], 0, 0);

  /* increment the system memory address to the addr of the next chunk in the array */
  in_addr = in_addr + CHUNK_SIZE * sizeof(float);

  /* Issue the 2nd DMA */
  mfc_get (local_buffers[1], in_addr, CHUNK_SIZE * sizeof(float), tags[1], 0, 0);

  /* increment the system memory address to the addr of the next chunk in the array */
  in_addr = in_addr + CHUNK_SIZE * sizeof(float);

  /* wait for the first DMA to complete */
  mfc_write_tag_mask (1 << tags[0]);
  mfc_read_tag_status_all ();

  /* process local_buffers[buf0] */
#ifdef NO_SLICE 
	AUXILIARY(tag1 = tags[0]);
	AUXILIARY(tag2 = tags[1]);
	AUXILIARY(tag3 = tags[2]);	
	AUXILIARY(process_data_simd (local_buffers[0], CHUNK_SIZE));
	AUXILIARY(__CPROVER_assume(tag1 == tags[0]));
	AUXILIARY(__CPROVER_assume(tag2 == tags[1]));
	AUXILIARY(__CPROVER_assume(tag3 == tags[2]));
#endif

  /* set the buffer indices */
  put_buf = 0;
  process_buf = 1;
  get_buf = 2;

  for (i = 2; i < num_chunks; i++)
  {
#ifdef COMBINED_CASE
	  __CPROVER_k_induction_hint(3,3,1,0);
#endif
#ifdef SPLIT_CASE
	  __CPROVER_k_induction_hint(3,3,1,0);
#endif
	  
#ifndef ABS_INT
	  INVARIANT(assert(0 <= put_buf && put_buf <= 2));
	  INVARIANT(assert(0 <= process_buf && process_buf <= 2));
	  INVARIANT(assert(0 <= get_buf && get_buf <= 2));
#endif
	  
	  int tmp;

    // ASSERTIONS TO HELP SCRATCHY
    assert((put_buf >= 0) && (put_buf < 3));

    /* Issue a DMA put to move the buf0 to system mem */
    mfc_put (local_buffers[put_buf], out_addr, 
        CHUNK_SIZE * sizeof(float), tags[put_buf], 0, 0);

    /* update out_addr */
    out_addr += CHUNK_SIZE * sizeof(float);

    /* get the next input buffer */
    mfc_write_tag_mask(1 << tags[get_buf]);
    mfc_read_tag_status_all();
    mfc_get
    (local_buffers[get_buf], in_addr, CHUNK_SIZE * sizeof(float),
        tags[get_buf], 0, 0);
    in_addr += CHUNK_SIZE * sizeof(float);
   
    /* wait for the buffers local_buffers[buf1]  */
    mfc_write_tag_mask (1 << tags[process_buf]);
#ifdef CORRECT
    mfc_read_tag_status_all ();
#else
    /* BUG - wait has been removed */
#endif

    /* compute the current input buffer  */
#ifdef NO_SLICE 
	  AUXILIARY(the_process_buf = process_buf);
	  AUXILIARY(tag1 = tags[0]);
	  AUXILIARY(tag2 = tags[1]);
	  AUXILIARY(tag3 = tags[2]);
	  AUXILIARY(the_get_buf = get_buf);
	  AUXILIARY(the_put_buf = put_buf);
	  AUXILIARY(the_get_tag = tags[get_buf]);
	  AUXILIARY(the_put_tag = tags[put_buf]);
	  /* compute the current input buffer  */
#ifdef SPLIT_CASE
	  process_data_simd2
#else
	  process_data_simd
#endif
	   (local_buffers[process_buf], CHUNK_SIZE);
	  AUXILIARY(__CPROVER_assume(the_process_buf == process_buf));
	  AUXILIARY(__CPROVER_assume(tag1 == tags[0]));
	  AUXILIARY(__CPROVER_assume(tag2 == tags[1]));
	  AUXILIARY(__CPROVER_assume(tag3 == tags[2]));
	  AUXILIARY(__CPROVER_assume(the_get_buf == get_buf));
	  AUXILIARY(__CPROVER_assume(the_put_buf == put_buf));
	  AUXILIARY(__CPROVER_assume(the_get_tag == tags[get_buf]));
	  AUXILIARY(__CPROVER_assume(the_put_tag == tags[put_buf]));	  
#endif

    /* update buffer indices */
    tmp = put_buf;
    put_buf = process_buf;
    process_buf = get_buf;
    get_buf = tmp;
  }

  /* issue DMA put to move the cur_out_buf to system mem */
  mfc_put (local_buffers[put_buf], out_addr, 
      CHUNK_SIZE * sizeof(float), tags[put_buf], 0, 0);

  out_addr += CHUNK_SIZE * sizeof(float);

  /* wait for the last buffer to arrive */
  mfc_write_tag_mask (1 << tags[process_buf]);
  mfc_read_tag_status_all ();

  /* compute the current input buffer */
#ifdef NO_SLICE
    AUXILIARY(the_process_buf = process_buf);
    AUXILIARY(tag1 = tags[0]);
    AUXILIARY(tag2 = tags[1]);
    AUXILIARY(tag3 = tags[2]);	
    AUXILIARY(the_get_buf = get_buf);
    AUXILIARY(the_put_buf = put_buf);
    AUXILIARY(the_get_tag = tags[get_buf]);
    AUXILIARY(the_put_tag = tags[put_buf]);
	/* compute the current input buffer */
#ifdef SPLIT_CASE
	process_data_simd2
#else
    process_data_simd
#endif
	(local_buffers[process_buf], CHUNK_SIZE);
	
    AUXILIARY(__CPROVER_assume(the_process_buf == process_buf));
    AUXILIARY(__CPROVER_assume(tag1 == tags[0]));
    AUXILIARY(__CPROVER_assume(tag2 == tags[1]));
    AUXILIARY(__CPROVER_assume(tag3 == tags[2]));	
    AUXILIARY(__CPROVER_assume(the_get_tag == tags[get_buf]));
    AUXILIARY(__CPROVER_assume(the_put_tag == tags[put_buf]));
    AUXILIARY(__CPROVER_assume(the_get_buf == get_buf));
    AUXILIARY(__CPROVER_assume(the_put_buf == put_buf));
#endif
	
  put_buf = process_buf;

  /* DMA the last result to system memory */
  mfc_put (local_buffers[put_buf], out_addr, 
        CHUNK_SIZE * sizeof(float), tags[put_buf], 0, 0);

  /* wait for the last output to complete */
  mfc_write_tag_mask (1 << tags[put_buf]);
  mfc_read_tag_status_all ();

#ifdef USE_TIMER
  time_working = (spu_clock_read() - start);
  spu_clock_stop();
  printf ("SPE time_working = %lld\n", time_working);
#endif /* USE_TIMER */

  return 0;
}
