/* --------------------------------------------------------------  */
/* (C)Copyright 2001,2008,                                         */
/* International Business Machines Corporation,                    */
/* Sony Computer Entertainment, Incorporated,                      */
/* Toshiba Corporation,                                            */
/*                                                                 */
/* All Rights Reserved.                                            */
/*                                                                 */
/* Redistribution and use in source and binary forms, with or      */
/* without modification, are permitted provided that the           */
/* following conditions are met:                                   */
/*                                                                 */
/* - Redistributions of source code must retain the above copyright*/
/*   notice, this list of conditions and the following disclaimer. */
/*                                                                 */
/* - Redistributions in binary form must reproduce the above       */
/*   copyright notice, this list of conditions and the following   */
/*   disclaimer in the documentation and/or other materials        */
/*   provided with the distribution.                               */
/*                                                                 */
/* - Neither the name of IBM Corporation nor the names of its      */
/*   contributors may be used to endorse or promote products       */
/*   derived from this software without specific prior written     */
/*   permission.                                                   */
/*                                                                 */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND          */
/* CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,     */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF        */
/* MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE        */
/* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR            */
/* CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,    */
/* SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT    */
/* NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;    */
/* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)        */
/* HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN       */
/* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR    */
/* OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,  */
/* EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/* --------------------------------------------------------------  */
/* PROLOG END TAG zYx                                              */
#include <spu_intrinsics.h>
#include <spu_mfcio.h>
#include "particle.h"

#ifdef COMBINED_CASE
#define INVARIANT(X)
#define AUXILIARY(X)
#else
#define INVARIANT(X) X
#define AUXILIARY(X) X
#endif
#define REQUIRED_INVARIANT(X) X


#define PARTICLES_PER_BLOCK		1024	/* # of particles to process at a time */

// Local store structures and buffers.
volatile parm_context ctx;
volatile vec_float4 pos[PARTICLES_PER_BLOCK] __attribute__ ((aligned (128)));
volatile vec_float4 vel[PARTICLES_PER_BLOCK] __attribute__ ((aligned (128)));
volatile float inv_mass[PARTICLES_PER_BLOCK] __attribute__ ((aligned (128)));

int spu_main(unsigned long long spu_id __attribute__ ((unused)), unsigned long long parm)
{
  int i, j;
  int left, cnt;
  float time;
  unsigned int tag_id;
  vec_float4 dt_v, dt_inv_mass_v;

  // Reserve a tag ID
  tag_id = mfc_tag_reserve();

  mfc_write_tag_mask(-1);

  // Input parameter parm is a pointer to the particle parameter context.
  // Fetch the context, waiting for it to complete.
  
  mfc_get((void *)(&ctx), (unsigned int)parm, sizeof(parm_context), tag_id, 0, 0);
  HAVOC_CONTEXT(ctx);

  REQUIRED_INVARIANT(__CPROVER_assume(ctx.particles < 1000000));

  mfc_read_tag_status_all();

  //  dt_v = spu_splats(ctx.dt);

  // For each step in time
  for(time = 0; time < END_OF_TIME; time += ctx.dt) {
#ifdef COMBINED_CASE
	  __CPROVER_k_induction_hint(1,1,1,0);	  
#endif
#ifdef SPLIT_CASE
	  __CPROVER_k_induction_hint(1,1,1,0);	  
#endif
	  
	// For each block of particles
    for (i=0; i<ctx.particles; i+=PARTICLES_PER_BLOCK) {
#ifdef COMBINED_CASE
		__CPROVER_k_induction_hint(2,2,1,0);	  
#endif
#ifdef SPLIT_CASE
		__CPROVER_k_induction_hint(2,2,1,0);	  
#endif
		
		// Determine the number of particles in this block.
      left = ctx.particles - i;
		
	  INVARIANT(assert(0 <= i && i < ctx.particles));
	  INVARIANT(assert(0 <= ctx.particles && ctx.particles < 1000000));
		
      cnt = (left < PARTICLES_PER_BLOCK) ? left : PARTICLES_PER_BLOCK;

		// Fetch the data - position, velocity and inverse_mass. Wait for the DMA to complete 
	// before performing computation.
#ifdef CORRECT
	mfc_getb
#else
        /* BUG - get with barrier replaced by standard get - this causes a DMA race */
	mfc_get
#endif
	((void *)(pos), (unsigned int)(ctx.pos_v+i), cnt * sizeof(vec_float4), tag_id, 0, 0);
	mfc_get((void *)(vel), (unsigned int)(ctx.vel_v+i), cnt * sizeof(vec_float4), tag_id, 0, 0);
	mfc_get((void *)(inv_mass), (unsigned int)(ctx.inv_mass+i), cnt * sizeof(float), tag_id, 0, 0);

	INVARIANT(mfc_write_tag_mask(-1));

	mfc_read_tag_status_all();
	
	// Compute the step in time for the block of particles

#ifdef NO_SLICE
	for (j=0; j<cnt; j++) {
#ifdef COMBINED_CASE
		__CPROVER_k_induction_hint(0,0,1,0);	  
#endif
#ifdef SPLIT_CASE
		__CPROVER_k_induction_hint(1,1,1,0);	  
#endif
		INVARIANT(dma_free_region(pos, PARTICLES_PER_BLOCK*sizeof(vec_float4)));
		INVARIANT(dma_free_region(vel, PARTICLES_PER_BLOCK*sizeof(vec_float4)));
		INVARIANT(dma_free_region(inv_mass, PARTICLES_PER_BLOCK*sizeof(float)));
		INVARIANT(assert(tag_id < 32));
		INVARIANT(assert(0 <= ctx.particles && ctx.particles < 1000000));
		INVARIANT(assert(0 <= cnt && cnt <= PARTICLES_PER_BLOCK));
		INVARIANT(assert(0 <= i && i < ctx.particles));
#ifdef USE_SIMD
	  pos[j] = spu_madd(vel[j], dt_v, pos[j]);
	  dt_inv_mass_v = spu_mul(dt_v, spu_splats(inv_mass[j]));
	  vel[j] = spu_madd(dt_inv_mass_v, ctx.force_v, vel[j]);
#endif
	}
#endif
	
	// Put the position and velocity data back into system memory
	mfc_put((void *)(pos), (unsigned int)(ctx.pos_v+i), cnt * sizeof(vec_float4), tag_id, 0, 0);
	mfc_put((void *)(vel), (unsigned int)(ctx.vel_v+i), cnt * sizeof(vec_float4), tag_id, 0, 0);
    }
  }

  // Wait for final DMAs to complete before terminating SPU thread.
  mfc_read_tag_status_all();
  return (0);
}
